from pathlib import Path
from typing import Optional, Tuple

import pandas as pd
from aislib.misc_utils import ensure_path_exists
from skopt import Optimizer

from eir_auto_gp.single_task.modelling.feature_selection_utils import (
    gather_fractions_and_performances,
    read_gwas_df,
)
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


def run_gwas_bo_feature_selection(
    fold: int,
    folder_with_runs: Path,
    feature_selection_output_folder: Path,
    gwas_output_folder: Optional[Path],
    gwas_p_value_threshold: Optional[float],
) -> Optional[Path]:
    fs_out_folder = feature_selection_output_folder
    subsets_out_folder = fs_out_folder / "snp_importance" / "snp_subsets"
    snp_subset_file = subsets_out_folder / f"chosen_snps_{fold}.txt"

    if snp_subset_file.exists():
        return snp_subset_file

    fractions_file = subsets_out_folder / f"chosen_snps_fraction_{fold}.txt"

    assert gwas_output_folder is not None
    df_gwas = read_gwas_df(gwas_output_folder=gwas_output_folder)
    df_gwas = df_gwas.rename(columns={"P": "GWAS P-VALUE"})
    df_gwas = df_gwas[["GWAS P-VALUE"]]

    top_n, fraction = get_gwas_bo_auto_top_n(
        df_gwas=df_gwas,
        folder_with_runs=folder_with_runs,
        feature_selection_output_folder=feature_selection_output_folder,
        fold=fold,
        gwas_p_value_threshold=gwas_p_value_threshold,
    )
    logger.info("Top %d SNPs selected.", top_n)

    df_top_n = get_gwas_top_n_snp_list_df(df_gwas=df_gwas, top_n_snps=top_n)
    df_top_n_snps_only = df_top_n[["SNP"]]
    ensure_path_exists(path=snp_subset_file)
    df_top_n_snps_only.to_csv(path_or_buf=snp_subset_file, index=False, header=False)
    fractions_file.write_text(str(fraction))

    return snp_subset_file


def get_gwas_bo_auto_top_n(
    df_gwas: pd.DataFrame,
    folder_with_runs: Path,
    feature_selection_output_folder: Path,
    fold: int,
    gwas_p_value_threshold: Optional[float],
    min_n_snps: int = 16,
) -> Tuple[int, float]:
    max_fraction = _compute_max_fraction(
        df_gwas=df_gwas,
        gwas_p_value_threshold=gwas_p_value_threshold,
    )

    manual_fractions, manual_p_values = _get_manual_gwas_bo_fractions(
        df_gwas=df_gwas,
        min_snps_cutoff=1,
        max_fraction=max_fraction,
    )

    n_snps = len(df_gwas)

    if fold < len(manual_fractions):
        next_fraction = manual_fractions[fold]
        logger.info(
            "Next manual fraction for GWAS+BO: %f (p-value: %.2e)",
            next_fraction,
            manual_p_values[fold],
        )
    else:
        threshold_snps = min(min_n_snps, n_snps)
        min_fraction = threshold_snps / n_snps
        logger.debug("Setting minimum fraction to %.2e.", min_fraction)

        opt = Optimizer(
            dimensions=[(min_fraction, max_fraction, "log-uniform")],
            n_initial_points=len(manual_fractions),
        )
        df_history = gather_fractions_and_performances(
            folder_with_runs=folder_with_runs,
            feature_selection_output_folder=feature_selection_output_folder,
        )

        for t in df_history.itertuples():
            negated_performance = -t.best_val_performance
            opt.tell([t.fraction], negated_performance)

        next_fraction = opt.ask()[0]
        logger.info("Next computed fraction for GWAS+BO: %f", next_fraction)

    top_n = int(next_fraction * len(df_gwas))

    if top_n < min_n_snps:
        if n_snps >= min_n_snps:
            top_n = min_n_snps
            logger.info(
                "Computed top_n for GWAS+BO %d is too small (< %d). Setting to 16.",
                min_n_snps,
                top_n,
            )
        else:
            top_n = n_snps
            logger.info(
                "Dataset contains only %d SNPs, less than %d. Using all %d SNPs.",
                n_snps,
                min_n_snps,
                top_n,
            )

        next_fraction = top_n / n_snps

    return top_n, next_fraction


def _compute_max_fraction(
    df_gwas: pd.DataFrame, gwas_p_value_threshold: Optional[float]
) -> float:
    if gwas_p_value_threshold is None:
        return 1.0

    df_subset = df_gwas[df_gwas["GWAS P-VALUE"] < gwas_p_value_threshold].copy()
    n_snps = len(df_subset)
    fraction = n_snps / len(df_gwas)

    logger.info(
        "Computed max fraction of SNPs with p-value < %f: %f for GWAS+BO.",
        gwas_p_value_threshold,
        fraction,
    )

    assert fraction > 0.0
    return fraction


def _get_manual_gwas_bo_fractions(
    df_gwas: pd.DataFrame,
    min_snps_cutoff: int,
    max_fraction: float,
) -> tuple[list[float], list[float]]:
    p_values = []
    fractions = []
    for p in range(8, 2, -1):
        p_value = 10**-p
        df_subset = df_gwas[df_gwas["GWAS P-VALUE"] < p_value]
        n_snps = len(df_subset)

        if n_snps < min_snps_cutoff:
            logger.info(
                "Skipping p-value for GWAS+BO %f due to %d SNPs being too few (<%d).",
                p_value,
                n_snps,
                min_snps_cutoff,
            )
            continue

        fraction = n_snps / len(df_gwas)
        if fraction > max_fraction:
            logger.info(
                "Skipping p-value for GWAS+BO %f due to %d SNPs being too many (> %f).",
                p_value,
                n_snps,
                max_fraction,
            )
            continue

        fractions.append(fraction)
        p_values.append(p_value)

    return fractions, p_values


def get_gwas_top_n_snp_list_df(df_gwas: pd.DataFrame, top_n_snps: int) -> pd.DataFrame:
    df = df_gwas.sort_values(by="GWAS P-VALUE", ascending=True)
    df_top_n = df.iloc[:top_n_snps, :].copy()
    df_top_n.index.name = "SNP"
    df_top_n["SNP"] = df_top_n.index
    return df_top_n



from pathlib import Path
from shutil import copyfile
from typing import Generator, List, Optional

import luigi
import numpy as np
from plink_pipelines.make_dataset import (
    RenameOnFailureMixin,
    _get_one_hot_encoded_generator,
    get_sample_generator_from_bed,
)

from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class Config(luigi.Task, RenameOnFailureMixin):
    output_folder = luigi.Parameter()

    @property
    def input_name(self):
        bed_files = [
            i for i in Path(str(self.raw_data_path)).iterdir() if i.suffix == ".bed"
        ]
        if len(bed_files) != 1:
            raise ValueError(
                f"Expected one .bed file in {self.raw_data_path}, but"
                f"found {bed_files}."
            )
        return str(bed_files[0])

    @property
    def file_name(self):
        raise NotImplementedError

    def output_target(self, file_name: str):
        output_path = Path(str(self.output_folder), file_name)

        return luigi.LocalTarget(str(output_path))

    def output(self):
        return self.output_target(self.file_name)


def _get_plink_inputs_from_folder(folder_path: Path) -> List[Path]:
    files = [i.with_suffix("") for i in folder_path.iterdir() if i.suffix == ".bed"]

    return files


class ExternalRawData(luigi.ExternalTask):
    raw_data_path = luigi.Parameter()

    @property
    def input_name(self):
        bed_files = [
            i for i in Path(str(self.raw_data_path)).iterdir() if i.suffix == ".bed"
        ]
        if len(bed_files) != 1:
            raise ValueError(
                f"Expected one .bed file in {self.raw_data_path}, but"
                f"found {bed_files}."
            )
        return str(bed_files[0])

    def output(self):
        return luigi.LocalTarget(str(self.input_name))


def get_encoded_snp_stream(
    bed_path: Path,
    chunk_size: int,
) -> Generator[tuple[str, np.ndarray], None, None]:
    chunk_generator = get_sample_generator_from_bed(
        bed_path=bed_path,
        chunk_size=chunk_size,
    )

    yield from _get_one_hot_encoded_generator(chunked_sample_generator=chunk_generator)


def copy_bim_file(
    source_folder: Path,
    output_folder: Path,
    ensure_folder_exists: bool = True,
) -> Optional[Path]:
    bim_files = list(source_folder.glob("*.bim"))

    assert_msg = f"Expected one .bim file in {source_folder}, found {len(bim_files)}"
    assert len(bim_files) == 1, assert_msg

    bim_path = bim_files[0]
    output_path = output_folder / "data_final.bim"

    if ensure_folder_exists:
        output_path.parent.mkdir(parents=True, exist_ok=True)

    copyfile(src=bim_path, dst=output_path)
    return output_path



import argparse
from dataclasses import dataclass
from pathlib import Path
from shutil import copyfile
from typing import Generator

import numpy as np
import pandas as pd
from aislib.misc_utils import ensure_path_exists, get_logger

from eir_auto_gp.predict.data_preparation_utils import (
    log_overlap,
    read_bim_and_cast_dtypes,
)
from eir_auto_gp.preprocess.genotype import get_encoded_snp_stream

logger = get_logger(name=__name__)


@dataclass
class PlinkFileSet:
    bed: Path
    bim: Path
    fam: Path


def get_plink_fileset_from_folder(folder_path: Path) -> PlinkFileSet:
    bed_files = [i for i in folder_path.iterdir() if i.suffix == ".bed"]
    bim_files = [i for i in folder_path.iterdir() if i.suffix == ".bim"]
    fam_files = [i for i in folder_path.iterdir() if i.suffix == ".fam"]

    if len(bed_files) != 1:
        raise ValueError(
            f"Expected one .bed file in {folder_path}, but" f"found {bed_files}."
        )
    if len(bim_files) != 1:
        raise ValueError(
            f"Expected one .bim file in {folder_path}, but" f"found {bim_files}."
        )
    if len(fam_files) != 1:
        raise ValueError(
            f"Expected one .fam file in {folder_path}, but" f"found {fam_files}."
        )

    return PlinkFileSet(
        bed=bed_files[0],
        bim=bim_files[0],
        fam=fam_files[0],
    )


@dataclass
class PreparedData:
    array_folder: Path
    bim_file: Path
    log_file: Path


def run_prepare_data(
    genotype_data_path: Path,
    array_chunk_size: int,
    reference_bim_to_project_to: Path,
    output_folder: Path,
) -> PreparedData:

    plink_fileset = get_plink_fileset_from_folder(folder_path=genotype_data_path)

    ref_bim = reference_bim_to_project_to
    df_bim_reference = read_bim_and_cast_dtypes(bim_file_path=ref_bim)

    df_bim = read_bim_and_cast_dtypes(bim_file_path=plink_fileset.bim)

    log_output_path = output_folder / "snp_overlap_analysis.txt"
    log_overlap(
        df_bim_prd=df_bim,
        df_bim_exp=df_bim_reference,
        output_path=log_output_path,
    )

    snp_mapping = create_snp_mapping(from_bim=df_bim, to_reference_bim=df_bim_reference)

    from_stream = get_encoded_snp_stream(
        bed_path=plink_fileset.bed,
        chunk_size=array_chunk_size,
    )

    projected_stream = get_projected_snp_stream(
        from_stream=from_stream,
        snp_mapping=snp_mapping,
    )

    copyfile(src=ref_bim, dst=output_folder / "projected.bim")

    array_outfolder = output_folder / "encoded_arrays"
    ensure_path_exists(path=array_outfolder, is_folder=True)

    for sample_id, projected_array in projected_stream:
        np.save(file=array_outfolder / f"{sample_id}.npy", arr=projected_array)

    return PreparedData(
        array_folder=array_outfolder,
        bim_file=ref_bim,
        log_file=log_output_path,
    )


@dataclass
class SNPMapping:
    direct_match_source_indices: np.ndarray
    direct_match_target_indices: np.ndarray
    flip_match_source_indices: np.ndarray
    flip_match_target_indices: np.ndarray
    missing_target_indices: np.ndarray
    n_source_snps: int
    n_target_snps: int


def create_snp_mapping(
    from_bim: pd.DataFrame,
    to_reference_bim: pd.DataFrame,
) -> SNPMapping:
    """
    from_bim: DataFrame with SNPs from the dataset we want to project
    to_reference_bim: DataFrame with SNPs in the reference space we want to
    project into
    """
    from_bim["pos_key"] = from_bim["CHR_CODE"] + ":" + from_bim["BP_COORD"].astype(str)
    to_reference_bim["pos_key"] = (
        to_reference_bim["CHR_CODE"] + ":" + to_reference_bim["BP_COORD"].astype(str)
    )

    from_bim["allele_key"] = from_bim.apply(
        lambda x: tuple(sorted([x["REF"], x["ALT"]])), axis=1
    )
    to_reference_bim["allele_key"] = to_reference_bim.apply(
        lambda x: tuple(sorted([x["REF"], x["ALT"]])), axis=1
    )

    matches = pd.merge(
        from_bim.reset_index().rename(columns={"index": "source_idx"}),
        to_reference_bim.reset_index().rename(columns={"index": "target_idx"}),
        on="pos_key",
        suffixes=("_source", "_target"),
    )

    direct_mask = (matches["REF_source"] == matches["REF_target"]) & (
        matches["ALT_source"] == matches["ALT_target"]
    )
    flip_mask = (matches["REF_source"] == matches["ALT_target"]) & (
        matches["ALT_source"] == matches["REF_target"]
    )

    direct_matches = matches[direct_mask]
    direct_source_indices = direct_matches["source_idx"].to_numpy()
    direct_target_indices = direct_matches["target_idx"].to_numpy()

    flip_matches = matches[flip_mask]
    flip_source_indices = flip_matches["source_idx"].to_numpy()
    flip_target_indices = flip_matches["target_idx"].to_numpy()

    all_matched_targets = set(direct_target_indices) | set(flip_target_indices)
    missing_target_indices = np.array(
        [i for i in range(len(to_reference_bim)) if i not in all_matched_targets]
    )

    return SNPMapping(
        direct_match_source_indices=direct_source_indices,
        direct_match_target_indices=direct_target_indices,
        flip_match_source_indices=flip_source_indices,
        flip_match_target_indices=flip_target_indices,
        missing_target_indices=missing_target_indices,
        n_source_snps=len(from_bim),
        n_target_snps=len(to_reference_bim),
    )


def get_projected_snp_stream(
    from_stream: Generator[tuple[str, np.ndarray], None, None],
    snp_mapping: SNPMapping,
) -> Generator[tuple[str, np.ndarray], None, None]:
    for sample_id, source_array in from_stream:
        projected_array = np.zeros((4, snp_mapping.n_target_snps), dtype=np.uint8)
        projected_array[3, :] = 1

        projected_array[:, snp_mapping.direct_match_target_indices] = source_array[
            :, snp_mapping.direct_match_source_indices
        ]

        if len(snp_mapping.flip_match_source_indices) > 0:
            flipped_data = source_array[:, snp_mapping.flip_match_source_indices].copy()
            flipped_data[[0, 2]] = flipped_data[[2, 0]]
            projected_array[:, snp_mapping.flip_match_target_indices] = flipped_data

        # assert all columns sum to exactly 1
        assert np.allclose(projected_array.sum(axis=0), 1.0)
        yield sample_id, projected_array


def get_prepared_folder_path(output_folder: str) -> Path:
    return Path(output_folder, "processed/full_inds/full_chrs/encoded_outputs")


def run_prepare_wrapper(cl_args: argparse.Namespace) -> None:
    genotype_data_path = Path(cl_args.genotype_data_path)
    reference_bim_path = Path(cl_args.reference_bim_path)
    output_folder = Path(cl_args.output_folder)
    array_chunk_size = cl_args.array_chunk_size

    prepared_data = run_prepare_data(
        genotype_data_path=genotype_data_path,
        array_chunk_size=array_chunk_size,
        reference_bim_to_project_to=reference_bim_path,
        output_folder=output_folder,
    )

    logger.info(
        f"Data preparation completed. Results stored in: {prepared_data.array_folder}"
    )
    logger.info(f"BIM file: {prepared_data.bim_file}")
    logger.info(f"Log file: {prepared_data.log_file}")


def get_cl_args() -> argparse.Namespace:
    parser = get_parser()
    args = parser.parse_args()
    return args


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=(
            "Prepare genotype data by projecting it into a reference space and "
            "saving the resulting NumPy arrays."
        )
    )

    parser.add_argument(
        "--genotype_data_path",
        type=str,
        help="Path to the folder containing the genotype data to prepare "
        "(should contain .bed, .bim, and .fam files).",
        required=True,
    )

    parser.add_argument(
        "--reference_bim_path",
        type=str,
        help="Path to reference BIM file to project the genotype data into.",
        required=True,
    )

    parser.add_argument(
        "--output_folder",
        type=str,
        help="Folder where the prepared data will be stored.",
        required=True,
    )

    parser.add_argument(
        "--array_chunk_size",
        type=int,
        default=256,
        help="Number of SNPs to process in each chunk (default: 1000).",
    )

    return parser


def main() -> None:
    args = get_cl_args()
    run_prepare_wrapper(cl_args=args)


if __name__ == "__main__":
    main()



import warnings
from itertools import product
from pathlib import Path

import numpy as np
import pandas as pd
from aislib.misc_utils import get_logger
from eir.setup.input_setup_modules.setup_omics import read_bim
from joblib import Parallel, delayed
from statsmodels.base.model import LikelihoodModelResults
from statsmodels.tools.sm_exceptions import ConvergenceWarning

from eir_auto_gp.post_analysis.effect_analysis.genotype_effects import (
    get_snp_allele_maps,
    get_statsmodels_fit_function,
)

logger = get_logger(name=__name__)

warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
warnings.filterwarnings(action="ignore", category=RuntimeWarning)


def get_gxe_interaction_effects(
    df_inputs: pd.DataFrame,
    df_target: pd.DataFrame,
    df_genotype_missing: pd.DataFrame,
    bim_file: Path,
    target_type: str,
) -> pd.DataFrame:
    target_name = df_target.columns[0]
    df_combined = pd.concat([df_target, df_inputs], axis=1)

    df_bim = read_bim(bim_file_path=str(bim_file))
    snp_ids = [i for i in df_inputs.columns if not i.startswith("COVAR_")]
    allele_maps = get_snp_allele_maps(df_bim=df_bim, snp_ids=snp_ids)

    df_results = compute_interactions(
        df=df_combined,
        df_genotype_missing=df_genotype_missing,
        target_name=target_name,
        allele_maps=allele_maps,
        target_type=target_type,
    )

    return df_results


def compute_interactions(
    df: pd.DataFrame,
    df_genotype_missing: pd.DataFrame,
    target_name: str,
    allele_maps: dict[str, dict[str, str]],
    target_type: str,
) -> pd.DataFrame:
    logger.info(
        "Gathering GxE interaction allele effect results for '%s'.", target_name
    )

    snps = [i for i in df.columns if not i.startswith("COVAR_") and i != target_name]
    covar_columns = [i for i in df.columns if i.startswith("COVAR_")]

    parallel_worker = Parallel(n_jobs=-1)
    all_results = parallel_worker(
        delayed(_compute_interaction_effect_wrapper)(
            df=df,
            target_name=target_name,
            df_genotype_missing=df_genotype_missing,
            allele_maps=allele_maps,
            snp=snp,
            covar=covar,
            target_type=target_type,
            covar_columns=covar_columns,
        )
        for snp, covar in product(snps, covar_columns)
    )

    all_results = [i for i in all_results if i is not None]

    if len(all_results) == 0:
        return pd.DataFrame()

    return pd.concat(all_results)


def _compute_interaction_effect_wrapper(
    df: pd.DataFrame,
    target_name: str,
    df_genotype_missing: pd.DataFrame,
    snp: str,
    covar: str,
    target_type: str,
    allele_maps: dict[str, dict[str, str]],
    covar_columns: list[str],
) -> pd.DataFrame | None:

    other_covar_columns = [i for i in covar_columns if i != covar]

    formula = build_gxe_interaction_formula(
        target=target_name,
        snp=snp,
        covar=covar,
        other_covar_columns=other_covar_columns,
    )

    df_cur = df[[target_name, snp, *covar_columns]].dropna()
    mask = df_genotype_missing[snp] == 0
    df_cur_no_na = df_cur[mask]

    fit_func = get_statsmodels_fit_function(target_type=target_type)

    try:
        result = fit_func(formula=formula, data=df_cur_no_na).fit(disp=0)
        if hasattr(result, "mle_retvals"):
            if not result.mle_retvals["converged"]:
                raise RuntimeError("Model did not converge.")
        elif not np.all(np.isfinite(result.normalized_cov_params)):
            raise RuntimeError("Model did not converge.")

        result_summary = result.summary2().tables[1]
        if not np.all(np.isfinite(result_summary)):
            raise RuntimeError("Model resulted in non-finite values")

    except Exception as e:
        logger.error(
            f"Failed regression on '{snp}' and '{covar}' due to exception '{e}'. "
            f"Skipping."
        )
        return None

    df_result = build_df_from_interaction_results(
        results=result,
        allele_maps=allele_maps,
        snp=snp,
        covar=covar,
    )

    return df_result


def build_gxe_interaction_formula(
    target: str,
    snp: str,
    covar: str,
    other_covar_columns: list[str],
) -> str:
    formula = f"Q('{target}') ~ C(Q('{snp}')) + Q('{covar}') + Q('{snp}'):Q('{covar}')"

    for other_covar in other_covar_columns:
        formula += f" + Q('{other_covar}')"

    return formula


def build_df_from_interaction_results(
    results: LikelihoodModelResults,
    allele_maps: dict[str, dict[str, str]],
    snp: str,
    covar: str,
) -> pd.DataFrame:
    coef = results.params
    std_err = results.bse
    p_values = results.pvalues
    conf_int = results.conf_int()

    df = pd.DataFrame(
        {
            "Coefficient": coef,
            "STD ERR": std_err,
            "P>|t|": p_values,
            "0.025 CI": conf_int[0],
            "0.975 CI": conf_int[1],
        }
    )

    df.index.name = "Interaction"
    df_renamed = _rename_interaction_regression_index(
        df_results=df,
        allele_maps=allele_maps,
        snp=snp,
        covar=covar,
    )

    df_renamed = df_renamed.rename(index={f"{snp}:{covar}": f"{snp}--:--{covar}"})
    df_renamed["KEY"] = f"{snp}--:--{covar}"

    return df_renamed


def _rename_interaction_regression_index(
    df_results: pd.DataFrame,
    allele_maps: dict[str, dict[str, str]],
    snp: str,
    covar: str,
) -> pd.DataFrame:
    snp_allele_map = allele_maps[snp]
    cur_mapping = {}

    for index in df_results.index:
        if snp in index:
            if "C(Q(" in index:
                if "[T.1]" in index:
                    cur_allele = snp_allele_map["HET"]
                    new_index = f"{snp} {cur_allele}"
                elif "[T.2]" in index:
                    cur_allele = snp_allele_map["ALT"]
                    new_index = f"{snp} {cur_allele}"
                else:
                    cur_allele = snp_allele_map["REF"]
                    new_index = f"{snp} {cur_allele}"
                cur_mapping[index] = new_index

            elif ":Q('" in index:
                new_index = f"{snp}:{covar.strip()}"
                cur_mapping[index] = new_index

        elif "Q(" in index:
            start_pos = index.find("'") + 1
            end_pos = index.find("'", start_pos)
            covar_name = index[start_pos:end_pos]
            group_suffix = ""
            if "[T." in index and "]" in index:
                start = index.find("[T.") + 3
                end = index.find("]")
                group = index[start:end]

                group_suffix = f" (Group {group})"
            new_index = f"{covar_name}{group_suffix}"
            cur_mapping[index] = new_index

    df_results = df_results.rename(index=cur_mapping)
    return df_results



import argparse
import json
import logging
import re
import shutil
import subprocess
from copy import deepcopy
from pathlib import Path
from typing import Any

import pandas as pd
import torch
import yaml
from aislib.misc_utils import get_logger
from eir.train_utils.train_handlers import _iterdir_ignore_hidden

from eir_auto_gp.multi_task.modelling.run_modelling import (
    get_testing_string_from_config_folder,
)
from eir_auto_gp.predict.data_preparation_utils import get_experiment_bim_file
from eir_auto_gp.predict.pack import unpack_experiment
from eir_auto_gp.predict.prepare_data import run_prepare_data

logger = get_logger(name=__name__)

luigi_logger = logging.getLogger("luigi")
luigi_logger.setLevel(logging.INFO)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--genotype_data_path",
        type=str,
        help="The path to the genotype data to predict on.",
    )

    parser.add_argument(
        "--packed_experiment_path",
        type=str,
        help="Path to packed (zipped) experiment.",
    )
    parser.add_argument(
        "--output_folder",
        type=str,
        help="The folder to output the results.",
    )

    parser.add_argument(
        "--genotype_processing_chunk_size",
        type=int,
        default=256,
    )
    return parser


def run_sync_and_predict_wrapper(
    cl_args: argparse.Namespace,
) -> None:

    unpacked_experiment_path = unpack_experiment(
        packed_experiment_path=cl_args.packed_experiment_path,
        output_path=cl_args.output_folder + "/unpacked_experiment",
    )

    data_output_folder = Path(cl_args.output_folder, "data")
    experiment_folder = Path(cl_args.output_folder + "/unpacked_experiment")
    experiment_bim = get_experiment_bim_file(experiment_folder=experiment_folder)

    prepared_folder = run_prepare_data(
        genotype_data_path=Path(cl_args.genotype_data_path),
        output_folder=data_output_folder,
        reference_bim_to_project_to=experiment_bim,
        array_chunk_size=cl_args.genotype_processing_chunk_size,
    )

    run_predict(
        prepared_input_data_folder=prepared_folder.array_folder,
        unpacked_experiment_path=Path(unpacked_experiment_path),
        output_folder=Path(cl_args.output_folder),
    )

    gather_results(
        unpacked_experiment_path=Path(unpacked_experiment_path),
        output_folder=Path(cl_args.output_folder, "results"),
    )

    shutil.rmtree(path=data_output_folder)
    shutil.rmtree(path=unpacked_experiment_path)


def gather_results(unpacked_experiment_path: Path, output_folder: Path) -> None:
    output_folder.mkdir(parents=True, exist_ok=True)
    config_path = unpacked_experiment_path / "config.json"
    config = read_config(config_file=config_path)
    data_types = prepare_data_types(config=config)

    aggregated_data: dict[str, pd.DataFrame] = {}

    for predictions_file in unpacked_experiment_path.rglob("predictions.csv"):
        df = pd.read_csv(predictions_file)

        target_name = predictions_file.parent.name

        fold_path_part = predictions_file.parts[-5]
        assert fold_path_part.startswith("fold_")
        fold_number = fold_path_part.split("_")[-1]

        id_column = "ID"

        data_type = data_types[target_name]

        if data_type == "continuous":
            target_column = f"{target_name} Untransformed"
            new_column_name = f"{target_name} Fold {fold_number}"

            if target_column not in df.columns:
                logger.warning(
                    f"Column '{target_column}' not found in {predictions_file}"
                )
                continue

            df_selected = df[[id_column, target_column]].rename(
                columns={target_column: new_column_name}
            )
        elif data_type == "categorical":
            rename_map = {
                col: f"{target_name}: {col} Fold {fold_number}"
                for col in df.columns
                if col != id_column
            }
            df_selected = df.rename(columns=rename_map)
        else:
            raise ValueError(f"Unsupported data type: {data_type}")

        if target_name in aggregated_data:
            aggregated_data[target_name] = pd.merge(
                left=aggregated_data[target_name],
                right=df_selected,
                on="ID",
                how="outer",
            )
        else:
            aggregated_data[target_name] = df_selected

    for target_name, df_aggregated in aggregated_data.items():
        output_path = output_folder / f"{target_name}.csv"
        data_type = data_types[target_name]
        df_aggregated = compute_ensemble_and_uncertainty(
            df=df_aggregated,
            target=target_name,
            data_type=data_type,
        )
        df_aggregated.to_csv(output_path, index=False)


def read_config(config_file: Path) -> dict:
    with config_file.open("r") as file:
        config = json.load(file)
    return config


def prepare_data_types(config: dict) -> dict:
    data_types = {}
    for col in config["output_cat_columns"]:
        data_types[col] = "categorical"
    for col in config["output_con_columns"]:
        data_types[col] = "continuous"
    return data_types


def extract_categorical_class_name(column_name: str) -> str:
    class_name = re.sub(r"^.*?:", "", column_name).strip()
    return re.sub(r"\s+Fold\s+\d+$", "", class_name)


def compute_continuous_ensemble(
    df: pd.DataFrame,
    target_columns: list,
    target: str,
) -> pd.DataFrame:
    results = pd.DataFrame()
    results[f"{target} Ensemble"] = df[target_columns].mean(axis=1)

    bootstrap_lower_ci, bootstrap_upper_ci = [], []
    for row in df[target_columns].to_numpy():
        lower_ci, upper_ci = sample_bootstrap_ci(
            data=row,
            n_bootstraps=1000,
            ci=[2.5, 97.5],
        )
        bootstrap_lower_ci.append(lower_ci)
        bootstrap_upper_ci.append(upper_ci)

    results[f"{target} 2.5% CI"] = bootstrap_lower_ci
    results[f"{target} 97.5% CI"] = bootstrap_upper_ci
    return results


import numpy as np
import pandas as pd
from scipy.special import softmax


def compute_categorical_ensemble(
    *,
    df: pd.DataFrame,
    target_columns: list[str],
    target: str,
) -> pd.DataFrame:
    results = pd.DataFrame()
    classes = sorted(
        set(extract_categorical_class_name(column_name=col) for col in target_columns)
    )

    for class_name in classes:
        class_columns = [
            col
            for col in target_columns
            if extract_categorical_class_name(column_name=col) == class_name
        ]
        results[f"{target} Ensemble Raw {class_name}"] = df[class_columns].mean(axis=1)

    raw_outputs = results[[f"{target} Ensemble Raw {c}" for c in classes]].values

    if len(classes) == 1:
        class1_probs = 1 / (1 + np.exp(-raw_outputs))
        softmax_probs = np.hstack([1 - class1_probs, class1_probs])
        classes = ["0"] + classes
    else:
        softmax_probs = softmax(x=raw_outputs, axis=1)

    for i, class_name in enumerate(classes):
        results[f"{target} Ensemble Prob {class_name}"] = softmax_probs[:, i]

    epsilon = 1e-10
    entropies = -np.sum(softmax_probs * np.log(softmax_probs + epsilon), axis=1)
    max_entropy = np.log(len(classes))
    results[f"{target} Uncertainty"] = entropies / max_entropy

    results[f"{target} Predicted Class"] = [
        classes[i] for i in np.argmax(a=softmax_probs, axis=1)
    ]

    return results


def compute_ensemble_and_uncertainty(
    df: pd.DataFrame,
    target: str,
    data_type: str,
) -> pd.DataFrame:

    results = pd.DataFrame({"ID": df["ID"]})
    target_columns = [col for col in df.columns if col.startswith(target)]
    results = pd.concat([results, df[target_columns]], axis=1)

    if data_type == "continuous":
        ensemble_results = compute_continuous_ensemble(
            df=df,
            target_columns=target_columns,
            target=target,
        )
    elif data_type == "categorical":
        ensemble_results = compute_categorical_ensemble(
            df=df,
            target_columns=target_columns,
            target=target,
        )
    else:
        raise ValueError(f"Unsupported data type: {data_type}")

    results = pd.concat([results, ensemble_results], axis=1)
    return results


def sample_bootstrap_ci(data, n_bootstraps=1000, ci=(2.5, 97.5)):
    bootstrap_samples = np.random.choice(
        data,
        size=(n_bootstraps, data.shape[0]),
        replace=True,
    )
    bootstrap_means = np.mean(bootstrap_samples, axis=1)
    lower_ci, upper_ci = np.percentile(bootstrap_means, ci)
    return lower_ci, upper_ci


def run_predict(
    prepared_input_data_folder: Path,
    unpacked_experiment_path: Path,
    output_folder: Path,
) -> None:
    modelling_folder = unpacked_experiment_path / "modelling"
    assert modelling_folder.exists(), f"Folder {modelling_folder} does not exist."

    tmp_dir = output_folder / "tmp_configs"
    tmp_dir.mkdir(exist_ok=True, parents=True)
    for model_folder in _iterdir_ignore_hidden(path=modelling_folder):

        build_predict_configs(
            modelling_folder=model_folder,
            prepared_input_data_folder=prepared_input_data_folder,
            output_folder=tmp_dir,
        )

        command = get_testing_string_from_config_folder(
            config_folder=tmp_dir,
            train_run_folder=model_folder,
            with_labels=False,
        )
        logger.info(f"Running command: {command}")

        subprocess.run(command, shell=True)

    shutil.rmtree(path=tmp_dir)


def build_predict_configs(
    modelling_folder: Path,
    prepared_input_data_folder: Path,
    output_folder: Path,
) -> dict[str, Path]:
    configs_folder = modelling_folder / "configs"

    subset_folder = modelling_folder / "snp_subset_files"

    modified_files = {"output_config": []}

    for f in _iterdir_ignore_hidden(path=configs_folder):
        configs_as_dict = yaml.safe_load(f.read_text())
        match f.stem:
            case "global_config":
                config_device = configs_as_dict["basic_experiment"]["device"]
                if "cuda" in config_device and not torch.cuda.is_available():
                    logger.info(
                        "CUDA was used for original experiments but is not "
                        "available on current system. Changing device to CPU."
                    )
                    configs_as_dict["basic_experiment"]["device"] = "cpu"

                logger.info("Setting dataloader workers to 0 for prediction.")
                configs_as_dict["dataloader_workers"] = 0

                cur_config = configs_as_dict
                new_path = output_folder / f.name
                new_path.write_text(yaml.dump(cur_config))
                modified_files["global_config"] = new_path

            case "input_configs":
                for input_config in configs_as_dict:
                    cur_name = input_config["input_info"]["input_name"]

                    input_config["input_info"]["input_source"] = str(
                        prepared_input_data_folder
                    )

                    input_config = maybe_add_subset_to_input_config(
                        input_config=input_config,
                        subset_folder=subset_folder,
                        modelling_folder=modelling_folder,
                    )

                    new_path = output_folder / f"{cur_name}_input_config.yaml"
                    new_path.write_text(yaml.dump(input_config))
                    modified_files[cur_name] = new_path

            case "fusion_config":
                cur_config = configs_as_dict
                new_path = output_folder / f.name
                new_path.write_text(yaml.dump(cur_config))
                modified_files["fusion_config"] = new_path

            case "output_configs":
                for output_config in configs_as_dict:
                    cur_name = output_config["output_info"]["output_name"]
                    output_config["output_info"]["output_source"] = None
                    new_path = output_folder / f"{cur_name}_output_config.yaml"
                    new_path.write_text(yaml.dump(output_config))
                    modified_files["output_config"].append(new_path)

            case _:
                raise ValueError(f"File {f} is not recognized.")

    return modified_files


def maybe_add_subset_to_input_config(
    input_config: dict[str, Any],
    subset_folder: Path,
    modelling_folder: Path,
) -> dict[str, Any]:

    if not subset_folder.exists():
        return input_config

    input_config_copy = deepcopy(input_config)

    model_name = modelling_folder.name
    assert model_name.startswith("fold_")

    cur_fold = model_name.split("_")[-1]

    cur_subset_file_name = f"random_subset_fold={cur_fold}.txt"

    cur_subset_path = subset_folder / cur_subset_file_name
    assert cur_subset_path.exists(), f"Subset file {cur_subset_path} does not exist."

    logger.info(f"Adding subset file {cur_subset_path} to input config.")

    input_config_copy["input_type_info"]["subset_snps_file"] = str(cur_subset_path)

    return input_config_copy


def main() -> None:
    parser = get_parser()
    cl_args = parser.parse_args()
    run_sync_and_predict_wrapper(cl_args=cl_args)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Literal, Optional

from eir_auto_gp.single_task.modelling.dl_feature_selection import (
    run_dl_bo_selection,
    run_dl_plus_gwas_bo_selection,
)
from eir_auto_gp.single_task.modelling.gwas_bo_feature_selection import (
    run_gwas_bo_feature_selection,
)


def get_genotype_subset_snps_file(
    fold: int,
    folder_with_runs: Path,
    feature_selection_output_folder: Path,
    bim_file: str | Path,
    feature_selection_approach: Literal["dl", "gwas", "gwas->dl", "dl+gwas", None],
    n_dl_feature_selection_setup_folds: int,
    manual_subset_from_gwas: Optional[str | Path],
    gwas_p_value_threshold: Optional[float],
) -> Optional[Path]:
    match feature_selection_approach:
        case None:
            return None
        case "gwas":
            return manual_subset_from_gwas
        case "gwas+bo":
            assert manual_subset_from_gwas is not None
            computed_subset_file = run_gwas_bo_feature_selection(
                fold=fold,
                folder_with_runs=folder_with_runs,
                feature_selection_output_folder=feature_selection_output_folder,
                gwas_output_folder=Path(manual_subset_from_gwas).parent,
                gwas_p_value_threshold=gwas_p_value_threshold,
            )

            return computed_subset_file
        case "dl" | "gwas->dl":
            if feature_selection_approach == "gwas->dl":
                assert manual_subset_from_gwas is not None

            computed_subset_file = run_dl_bo_selection(
                fold=fold,
                folder_with_runs=folder_with_runs,
                feature_selection_output_folder=feature_selection_output_folder,
                bim_file=bim_file,
                n_dl_feature_selection_setup_folds=n_dl_feature_selection_setup_folds,
                manual_subset_from_gwas=manual_subset_from_gwas,
            )
            return computed_subset_file
        case "dl+gwas":
            if fold > n_dl_feature_selection_setup_folds:
                assert manual_subset_from_gwas is not None

            gwas_output_folder = manual_subset_from_gwas.parent
            computed_subset_file = run_dl_plus_gwas_bo_selection(
                fold=fold,
                folder_with_runs=folder_with_runs,
                feature_selection_output_folder=feature_selection_output_folder,
                gwas_output_folder=gwas_output_folder,
                bim_file=bim_file,
                n_dl_feature_selection_setup_folds=n_dl_feature_selection_setup_folds,
            )
            return computed_subset_file

        case _:
            raise ValueError()



from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.formula.api as smf
from aislib.misc_utils import ensure_path_exists, get_logger

from eir_auto_gp.post_analysis.effect_analysis.interaction_effects import (
    get_snp_allele_maps,
    read_bim,
)

logger = get_logger(name=__name__)


def run_grouped_interaction_analysis(
    df_genotype: pd.DataFrame,
    df_target: pd.DataFrame,
    df_interaction_effects: pd.DataFrame,
    top_n_snps: int,
    bim_file: Path,
    output_folder: Path,
) -> None:
    df_bim = read_bim(bim_file_path=str(bim_file))
    snp_ids = [i for i in df_genotype.columns if not i.startswith("COVAR_")]
    allele_maps = get_snp_allele_maps(df_bim=df_bim, snp_ids=snp_ids)

    df_genotype_prepared = prepare_genotype_data(
        df=df_genotype,
        allele_maps=allele_maps,
    )

    assert len(df_target.columns) == 1
    target_name = df_target.columns[0]

    df_combined = pd.concat(objs=[df_target, df_genotype_prepared], axis=1)

    all_snp_pairs = _get_snp_pairs_to_check(
        df_interaction_effects=df_interaction_effects,
        top_n=None,
    )

    snps_to_plot = _get_snp_pairs_to_check(
        df_interaction_effects=df_interaction_effects,
        top_n=top_n_snps,
    )

    all_results = []

    for snp1, snp2 in all_snp_pairs:
        df_results = fit_models_for_combinations(
            df=df_combined,
            snp1=snp1,
            snp2=snp2,
            target_name=target_name,
            allele_maps=allele_maps,
        )

        if [snp1, snp2] in snps_to_plot:
            output_path = output_folder / "figures" / f"{snp1}_{snp2}_interaction.pdf"
            ensure_path_exists(path=output_path)
            fig = plot_snp_coefficients_as_points(
                df_results=df_results,
                allele_maps=allele_maps,
                snp1_name=snp1,
                snp2_name=snp2,
            )

            fig.savefig(output_path)

        df_results = df_results.rename(
            columns={
                f"{snp1}_genotype": "SNP1_genotype",
                f"{snp2}_genotype": "SNP2_genotype",
            }
        )

        all_results.append(df_results)

    if not all_results:
        logger.warning("No significant SNP pairs found for grouped analysis.")
        return

    df_all_results = pd.concat(all_results)

    df_all_results.to_csv(
        output_folder / "snp_interactions_as_groups.csv",
        index=False,
    )

    all_snps = set([snp for pair in all_snp_pairs for snp in pair])
    snp_frequencies = []

    for snp in all_snps:
        cur_freq = calculate_snp_genotype_frequencies(
            df=df_genotype_prepared,
            snp=snp,
            genotype_maps=allele_maps,
        )
        snp_frequencies.append(cur_freq)

    df_snp_frequencies = pd.concat(snp_frequencies)
    df_snp_frequencies.to_csv(
        output_folder / "snp_frequencies.csv",
        index=False,
    )


def _get_snp_pairs_to_check(
    df_interaction_effects: pd.DataFrame,
    top_n: Optional[int] = None,
    p_value_threshold: float | str = "auto",
) -> list[list[str]]:
    if p_value_threshold == "auto":
        num_tests = df_interaction_effects["KEY"].nunique()
        p_value_threshold = 0.05 / num_tests
    elif not isinstance(p_value_threshold, float):
        raise ValueError("p_value_threshold must be 'auto' or a float.")

    logger.info(
        "Using p-value threshold of %f for grouped analysis.", p_value_threshold
    )

    interaction_keys = df_interaction_effects["KEY"].unique()
    df_all = df_interaction_effects
    df_ie = df_all[df_all.index.isin(interaction_keys)].copy()

    df_ie_filtered = df_ie[df_ie["P>|t|"] <= p_value_threshold]
    df_ie_interaction_keys = df_ie_filtered["KEY"].unique()
    assert len(df_ie_interaction_keys) == len(df_ie_filtered)

    df_to_check = df_ie_filtered.copy()

    if top_n is not None:
        df_to_check = df_ie_filtered.sort_values(
            "Coefficient",
            ascending=False,
        ).head(top_n)

    logger.info("Checking %d SNP pairs for grouped analysis.", len(df_to_check))

    return df_to_check["KEY"].str.split("--:--", expand=True).to_numpy().tolist()


def prepare_genotype_data(
    df: pd.DataFrame, allele_maps: dict[str, dict[str, str]]
) -> pd.DataFrame:
    df_prepared = df.copy()

    for snp, mappings in allele_maps.items():
        if snp in df.columns:
            df_prepared[snp] = df[snp].map(
                {
                    0: mappings["REF"],
                    1: mappings["HET"],
                    2: mappings["ALT"],
                }
            )
    return df_prepared


def fit_models_for_combinations(
    df: pd.DataFrame,
    snp1: str,
    snp2: str,
    target_name: str,
    allele_maps: dict[str, dict[str, str]],
) -> pd.DataFrame:
    results = []
    allele_order = ["REF", "HET", "ALT"]
    total_samples = len(df)

    for allele1 in allele_order:
        genotype1 = allele_maps[snp1].get(allele1)
        if genotype1 is None:
            continue

        for allele2 in allele_order:
            genotype2 = allele_maps[snp2].get(allele2)
            if genotype2 is None:
                continue

            subset = df[(df[snp1] == genotype1) & (df[snp2] == genotype2)]

            if not subset.empty:
                model = smf.ols(f"{target_name} ~ 1", data=subset).fit()
                conf_int = model.conf_int().loc["Intercept"]
                p_value = model.pvalues["Intercept"]
                combination_freq = len(subset) / total_samples
                results.append(
                    {
                        f"{snp1}_genotype": genotype1,
                        f"{snp2}_genotype": genotype2,
                        "Value": model.params["Intercept"],
                        "CI_lower": conf_int[0],
                        "CI_upper": conf_int[1],
                        "P_value": p_value,
                        "KEY": f"{snp1}--:--{snp2}",
                        "n": len(subset),
                        "Combination_freq": combination_freq,
                    }
                )
            else:
                logger.warning(
                    "No data for SNP1: %s, SNP2: %s, genotype1: %s, genotype2: %s",
                    snp1,
                    snp2,
                    genotype1,
                    genotype2,
                )

    return pd.DataFrame(results)


def plot_snp_coefficients_as_points(
    df_results: pd.DataFrame,
    allele_maps: dict[str, dict[str, str]],
    snp1_name: str,
    snp2_name: str,
) -> plt.Figure:
    sns.set_style("white")

    snp1_map = allele_maps[snp1_name]
    snp2_map = allele_maps[snp2_name]
    genotype_order_snp1 = [snp1_map["REF"], snp1_map["HET"], snp1_map["ALT"]]
    genotype_order_snp2 = [snp2_map["REF"], snp2_map["HET"], snp2_map["ALT"]]

    plt.rcParams.update(
        {
            "font.size": 14,
            "axes.titlesize": 16,
            "axes.labelsize": 14,
            "xtick.labelsize": 12,
            "ytick.labelsize": 12,
            "legend.fontsize": 12,
        }
    )
    fig, ax = plt.subplots(figsize=(8, 7))

    colors = ["#E69F00", "#56B4E9", "#009E73"]
    color_map = dict(zip(genotype_order_snp2, colors))

    for i, genotype1 in enumerate(genotype_order_snp1):
        for j, genotype2 in enumerate(genotype_order_snp2):
            subset = df_results[
                (df_results[f"{snp1_name}_genotype"] == genotype1)
                & (df_results[f"{snp2_name}_genotype"] == genotype2)
            ]

            if not subset.empty:
                row = subset.iloc[0]
                y = row["Value"]
                ci_lower = row["CI_lower"]
                ci_upper = row["CI_upper"]
                x = i + j * 0.1
                ax.errorbar(
                    x,
                    y,
                    yerr=[[y - ci_lower], [ci_upper - y]],
                    fmt="o",
                    markersize=8,
                    color=color_map[genotype2],
                    ecolor=color_map[genotype2],
                    elinewidth=2,
                    capsize=5,
                    capthick=2,
                    label=genotype2 if i == 0 else "",
                )

    ax.set_xticks(np.arange(len(genotype_order_snp1)) + 0.1)
    ax.set_xticklabels(genotype_order_snp1)

    ax.set_xlabel(snp1_name + " Genotype")
    ax.set_ylabel("Value")
    ax.set_title(f"Effect Sizes for {snp1_name} and {snp2_name} Genotypes")

    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    legend = ax.legend(
        by_label.values(),
        by_label.keys(),
        title=f"{snp2_name} Genotype",
        loc="upper left",
        frameon=False,
    )

    plt.setp(legend.get_title(), fontsize="13")

    sns.despine()

    plt.tight_layout()

    return fig


def calculate_snp_genotype_frequencies(
    df: pd.DataFrame,
    snp: str,
    genotype_maps: dict[str, dict[str, str]],
) -> pd.DataFrame:
    results = []
    total_samples = len(df)

    for genotype_label, genotype in genotype_maps[snp].items():
        count = df[snp].value_counts(dropna=False).get(genotype, 0)
        frequency = count / total_samples if total_samples > 0 else 0
        results.append(
            {
                "SNP": snp,
                "Genotype": genotype,
                "Genotype_Label": genotype_label,
                "Count": count,
                "Frequency": frequency,
            }
        )

    all_genotypes = set(df[snp].unique())
    known_genotypes = set(genotype_maps[snp].values())
    additional_genotypes = all_genotypes - known_genotypes - {float("nan")}

    for genotype in additional_genotypes:
        count = df[snp].value_counts(dropna=False).get(genotype, 0)
        frequency = count / total_samples if total_samples > 0 else 0
        results.append(
            {
                "SNP": snp,
                "Genotype": genotype,
                "Genotype_Label": "Other/Unknown",
                "Count": count,
                "Frequency": frequency,
            }
        )

    missing_count = df[snp].isna().sum()
    if missing_count > 0:
        results.append(
            {
                "SNP": snp,
                "Genotype": "Missing",
                "Genotype_Label": "Missing",
                "Count": missing_count,
                "Frequency": missing_count / total_samples,
            }
        )

    return pd.DataFrame(results)



from typing import TYPE_CHECKING

import pandas as pd
from aislib.misc_utils import ensure_path_exists, get_logger

from eir_auto_gp.post_analysis.effect_analysis.genotype_effects import (
    get_allele_effects,
)
from eir_auto_gp.post_analysis.effect_analysis.gxe_interaction_effects import (
    get_gxe_interaction_effects,
)
from eir_auto_gp.post_analysis.effect_analysis.interaction_effects import (
    get_interaction_effects,
)
from eir_auto_gp.post_analysis.effect_analysis.viz_genotype_effects import plot_top_snps
from eir_auto_gp.post_analysis.effect_analysis.viz_interaction_effects_graph import (
    generate_interaction_snp_graph_figure,
)
from eir_auto_gp.post_analysis.effect_analysis.viz_interaction_effects_point import (
    run_grouped_interaction_analysis,
)
from eir_auto_gp.post_analysis.run_complexity_analysis import (
    ModelReadyObject,
    convert_split_data_to_model_ready_object,
)

if TYPE_CHECKING:
    from eir_auto_gp.post_analysis.run_post_analysis import (
        PostAnalysisObject,
        SplitModelData,
    )

logger = get_logger(name=__name__)


def run_effect_analysis(post_analysis_object: "PostAnalysisObject") -> None:
    mro_genotype = convert_split_data_to_model_ready_object(
        split_model_data=post_analysis_object.modelling_data,
        include_genotype=True,
        include_tabular=True,
        one_hot_encode=False,
    )

    input_cat_columns = post_analysis_object.experiment_info.input_cat_columns

    df_inputs, df_target, df_genotype_missing = _build_effect_inputs(
        model_ready_object=mro_genotype,
        modelling_data=post_analysis_object.modelling_data,
        sets_for_effect_analysis=post_analysis_object.sets_for_effect_analysis,
        input_cat_columns=input_cat_columns,
    )

    pao = post_analysis_object

    output_root = pao.data_paths.analysis_output_path / "effect_analysis"
    ensure_path_exists(path=output_root, is_folder=True)

    effects_output = output_root / "allele_effects"
    ensure_path_exists(path=effects_output, is_folder=True)
    df_allele_effects = get_allele_effects(
        df_inputs=df_inputs,
        df_target=df_target,
        df_genotype_missing=df_genotype_missing,
        bim_file=pao.data_paths.snp_bim_path,
        target_type=pao.experiment_info.target_type,
    )
    df_allele_effects.to_csv(effects_output / "allele_effects.csv")

    df_allele_effects_only = filter_snp_rows(df=df_allele_effects)

    plot_top_snps(
        df=df_allele_effects_only,
        p_value_threshold=0.05,
        top_n=pao.top_n_genotype_snps_effects_to_plot,
        output_dir=effects_output / "figures",
    )

    interaction_output = output_root / "interaction_effects"
    ensure_path_exists(path=interaction_output, is_folder=True)
    df_interaction_effects = get_interaction_effects(
        df_inputs=df_inputs,
        df_target=df_target,
        df_genotype_missing=df_genotype_missing,
        bim_file=pao.data_paths.snp_bim_path,
        target_type=pao.experiment_info.target_type,
        allow_within_chr_interaction=pao.allow_within_chr_interaction,
        min_interaction_pair_distance=pao.min_interaction_pair_distance,
    )
    df_interaction_effects.to_csv(interaction_output / "interaction_effects.csv")

    if len(df_interaction_effects) > 0:

        df_interaction_effects_only = filter_snp_rows(df=df_interaction_effects)

        trait_name = df_target.columns[0]
        generate_interaction_snp_graph_figure(
            df_interaction_effects=df_interaction_effects_only,
            bim_file_path=pao.data_paths.snp_bim_path,
            df_target=df_target,
            trait=trait_name,
            plot_output_root=interaction_output,
            top_n_snps=pao.top_n_interaction_pairs,
        )

        run_grouped_interaction_analysis(
            df_genotype=df_inputs,
            df_target=df_target,
            df_interaction_effects=df_interaction_effects_only,
            top_n_snps=pao.top_n_interaction_pairs,
            bim_file=pao.data_paths.snp_bim_path,
            output_folder=output_root / "grouped_interaction_analysis",
        )

    gxe_output = output_root / "gxe_interaction_effects"
    ensure_path_exists(path=gxe_output, is_folder=True)
    df_gxe_interaction_effects = get_gxe_interaction_effects(
        df_inputs=df_inputs,
        df_target=df_target,
        df_genotype_missing=df_genotype_missing,
        bim_file=pao.data_paths.snp_bim_path,
        target_type=pao.experiment_info.target_type,
    )

    if len(df_gxe_interaction_effects) > 0:
        df_gxe_interaction_effects.to_csv(gxe_output / "gxe_interaction_effects.csv")


def _build_effect_inputs(
    model_ready_object: ModelReadyObject,
    modelling_data: "SplitModelData",
    sets_for_effect_analysis: list[str],
    input_cat_columns: list[str],
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Note we drop the first dummy column for each categorical column to avoid
    issues with multicollinearity in the linear models.
    """
    valid_sets = {"train", "valid", "test"}
    if not all(set_name in valid_sets for set_name in sets_for_effect_analysis):
        raise ValueError(
            "Invalid set names in sets_for_effect_analysis. "
            "Valid options are 'train', 'valid', 'test'."
        )

    input_dfs = []
    target_dfs = []
    missing_info_dfs = []

    if "train" in sets_for_effect_analysis:
        input_dfs.append(model_ready_object.input_train)
        target_dfs.append(model_ready_object.target_train)
        missing_info_dfs.append(modelling_data.train.df_genotype_nan_mask)
    if "valid" in sets_for_effect_analysis:
        input_dfs.append(model_ready_object.input_val)
        target_dfs.append(model_ready_object.target_val)
        missing_info_dfs.append(modelling_data.val.df_genotype_nan_mask)
    if "test" in sets_for_effect_analysis:
        input_dfs.append(model_ready_object.input_test)
        target_dfs.append(model_ready_object.target_test)
        missing_info_dfs.append(modelling_data.test.df_genotype_nan_mask)

    concatenated_input = pd.concat(input_dfs, ignore_index=True)
    concatenated_target = pd.concat(target_dfs, ignore_index=True)
    concatenated_missing = pd.concat(missing_info_dfs, ignore_index=True)

    for cat_column in input_cat_columns:
        dummy_columns = [
            col
            for col in concatenated_input.columns
            if col.startswith("COVAR_" + cat_column + "_")
        ]
        if dummy_columns:
            dummy_columns_sorted = sorted(dummy_columns)
            concatenated_input = concatenated_input.drop(
                columns=dummy_columns_sorted[0]
            )

    return concatenated_input, concatenated_target, concatenated_missing


def filter_snp_rows(df: pd.DataFrame) -> pd.DataFrame:
    df_filtered = df[~df.index.str.contains("COVAR")].copy()
    return df_filtered



import math
import os
import warnings
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Generator, Optional, Sequence, Tuple

import numpy as np

warnings.filterwarnings("ignore", message=".*newer version of deeplake.*")

import deeplake
import eir.data_load.label_setup
import luigi
import pandas as pd
from aislib.misc_utils import ensure_path_exists

from eir_auto_gp.preprocess.genotype import (
    ExternalRawData,
    copy_bim_file,
    get_encoded_snp_stream,
)
from eir_auto_gp.preprocess.tabular import ParseLabelFile
from eir_auto_gp.utils.utils import get_logger

logger = get_logger(name=__name__)


class CommonSplitIntoTestSet(luigi.Task):
    genotype_data_path = luigi.Parameter()
    label_file_path = luigi.Parameter()
    data_output_folder = luigi.Parameter()
    data_storage_format = luigi.Parameter()
    output_name = luigi.Parameter()
    pre_split_folder = luigi.Parameter()
    freeze_validation_set = luigi.BoolParameter()
    only_data = luigi.BoolParameter()
    genotype_processing_chunk_size = luigi.IntParameter()
    modelling_data_format = luigi.Parameter(default="disk")

    def requires(self):
        """
        Will require 2 tasks:
            - raw PLINK1 genotype data,
            - label file processing.

        Will gather IDs from all of these, then generate a split from that.

        One option for configuration is to allow custom overwriting of QC command.
        """

        geno_task = ExternalRawData(raw_data_path=self.genotype_data_path)

        label_file_task = ParseLabelFile(
            output_folder=self.data_output_folder,
            label_file_path=self.label_file_path,
        )
        return {
            "genotype": geno_task,
            "label_file": label_file_task,
        }

    def run(self):
        inputs = self.input()

        genotype_path = self.genotype_data_path
        bed_path = next(Path(str(genotype_path)).rglob("*.bed"), None)
        assert bed_path.exists()
        label_file_path = Path(str(inputs["label_file"].path))

        fam_path = next(Path(str(genotype_path)).rglob("*.fam"), None)
        assert fam_path.exists()

        output_root = Path(str(self.data_output_folder))
        genotype_output_folder = output_root / "genotype"
        bim_target_folder = genotype_output_folder / "processed/parsed_files/"
        copy_bim_file(
            source_folder=Path(str(genotype_path)),
            output_folder=bim_target_folder,
        )

        common_ids_to_keep = _gather_all_ids(
            fam_path=fam_path,
            label_file=label_file_path,
        )

        train_ids, valid_ids, test_ids = _id_setup_wrapper(
            output_root=output_root,
            pre_split_folder=self.pre_split_folder,
            common_ids_to_keep=common_ids_to_keep,
            freeze_validation_set=self.freeze_validation_set,
        )
        train_and_valid_ids = train_ids + valid_ids

        _split_csv_into_train_and_test(
            train_ids=train_and_valid_ids,
            test_ids=test_ids,
            source=label_file_path,
            destination=output_root / "tabular" / "final",
        )

        one_hot_stream = get_encoded_snp_stream(
            bed_path=bed_path,
            chunk_size=int(self.genotype_processing_chunk_size),
        )

        if self.data_storage_format == "deeplake":
            build_deeplake_train_and_test_ds_from_stream(
                train_ids=train_and_valid_ids,
                test_ids=test_ids,
                destination=output_root / "genotype" / "final",
                source_stream=one_hot_stream,
                commit_frequency=int(self.genotype_processing_chunk_size),
                chunk_size=int(self.genotype_processing_chunk_size),
            )
        elif self.data_storage_format == "disk":
            _build_disk_array_folders_from_stream(
                train_ids=train_and_valid_ids,
                test_ids=test_ids,
                source_stream=one_hot_stream,
                destination=output_root / "genotype" / "final",
            )
        else:
            raise ValueError(f"Unknown output format: {self.data_storage_format}")

    def output(self):
        output_root = Path(str(self.data_output_folder))
        outputs = {}

        for split in ("train", "test"):
            cur_target = luigi.LocalTarget(str(output_root / "genotype/final" / split))
            outputs[f"{split}_genotype"] = cur_target

            cur_tabular_path = output_root / "tabular/final" / f"labels_{split}"
            cur_tabular_path = cur_tabular_path.with_suffix(".csv")
            cur_tabular_target = luigi.LocalTarget(str(cur_tabular_path))
            outputs[f"{split}_tabular"] = cur_tabular_target

        return outputs


def get_genotype_path(task_input_path: str, data_storage_format: str) -> Path:
    if data_storage_format == "disk":
        return Path(task_input_path)
    elif data_storage_format == "deeplake":
        return Path(task_input_path) / "genotype"
    else:
        raise ValueError(f"Unknown output format: {data_storage_format}")


def _id_setup_wrapper(
    output_root: Path,
    common_ids_to_keep: Sequence[str],
    freeze_validation_set: bool,
    pre_split_folder: Optional[str] = None,
    check_valid_and_test_ids: bool = True,
) -> tuple[list[str], list[str], list[str]]:
    valid_path = None
    valid_path_exists = False

    if pre_split_folder is not None:
        pre_split_folder = Path(pre_split_folder)
        logger.info("Using pre-split folder: %s", pre_split_folder)

        train_path = pre_split_folder / "train_ids.txt"
        if not train_path.exists():
            raise FileNotFoundError(f"Could not find train IDs file: {train_path}")

        test_path = pre_split_folder / "test_ids.txt"
        if not test_path.exists():
            raise FileNotFoundError(f"Could not find test IDs file: {test_path}")

        train_ids = (
            pd.read_csv(train_path, header=None).astype(str).squeeze("columns").tolist()
        )
        check_extra_ids(
            ids_to_check=train_ids,
            id_set_name="train",
            common_ids=common_ids_to_keep,
        )

        test_ids = (
            pd.read_csv(test_path, header=None).astype(str).squeeze("columns").tolist()
        )
        if check_valid_and_test_ids:
            check_extra_ids(
                ids_to_check=test_ids,
                id_set_name="test",
                common_ids=common_ids_to_keep,
            )

        valid_path = pre_split_folder / "valid_ids.txt"
        valid_path_exists = valid_path.exists()

    else:
        logger.info("Generating train+valid and test IDs.")

        train_ids, test_ids = _split_ids(all_ids=common_ids_to_keep)

    valid_ids = []
    if freeze_validation_set:
        if not valid_path_exists:
            logger.info("Creating new frozen validation set.")

            batch_size = get_batch_size(samples_per_epoch=len(train_ids))
            valid_size = get_dynamic_valid_size(
                num_samples_per_epoch=len(train_ids),
                minimum=batch_size,
            )
            train_ids, valid_ids = _split_ids(
                all_ids=train_ids, valid_or_test_size=valid_size
            )
        else:
            logger.info("Using frozen validation set from %s.", valid_path)
            valid_ids = (
                pd.read_csv(valid_path, header=None)
                .astype(str)
                .squeeze("columns")
                .tolist()
            )
            if check_valid_and_test_ids:
                check_extra_ids(
                    ids_to_check=valid_ids,
                    id_set_name="valid",
                    common_ids=common_ids_to_keep,
                )

        _save_ids_to_text_file(
            ids=valid_ids, path=output_root / "ids" / "valid_ids.txt"
        )

    _save_ids_to_text_file(ids=train_ids, path=output_root / "ids" / "train_ids.txt")
    _save_ids_to_text_file(ids=test_ids, path=output_root / "ids" / "test_ids.txt")

    logger.info("Train and valid IDs: %d", len(train_ids) + len(valid_ids))
    logger.info("Test IDs: %d", len(test_ids))

    check_missing_ids(
        train_ids=train_ids,
        valid_ids=valid_ids,
        test_ids=test_ids,
        common_ids=common_ids_to_keep,
    )

    return train_ids, valid_ids, test_ids


def check_extra_ids(
    ids_to_check: Sequence[str],
    id_set_name: str,
    common_ids: Sequence[str],
    preview_limit: int = 10,
) -> None:
    common_ids_set = set(common_ids)

    extra_ids = list(set(ids_to_check) - common_ids_set)
    if extra_ids:
        preview = extra_ids[:preview_limit]
        raise ValueError(
            f"{id_set_name} contains {len(extra_ids)} IDs not in common IDs from "
            f"genotype data and label file."
            f"Preview of extra IDs: {preview}."
            f"Please check that the IDs in {id_set_name} "
            f"are available in both genotype data and label file."
        )


def check_missing_ids(
    train_ids: Sequence[str],
    valid_ids: Sequence[str],
    test_ids: Sequence[str],
    common_ids: Sequence[str],
    preview_limit: int = 10,
) -> None:
    common_ids_set = set(common_ids)

    combined_ids = set(train_ids) | set(valid_ids) | set(test_ids)
    missing_ids = list(common_ids_set - combined_ids)
    if missing_ids:
        preview = missing_ids[:preview_limit]
        logger.warning(
            f"Some common IDs are missing in the final sets. "
            f"Preview of missing IDs: {preview}"
        )


def _save_ids_to_text_file(ids: Sequence[str], path: Path) -> None:
    ensure_path_exists(path=path, is_folder=False)
    with open(path, "w") as f:
        for cur_id in ids:
            f.write(f"{cur_id}\n")


def _gather_all_ids(
    fam_path: Path,
    label_file: Path,
    filter_common: bool = True,
) -> Sequence[str]:
    fam_ids = gather_ids_from_fam(fam_path=fam_path)
    logger.info(
        "Gathered %d IDs from .fam file: %s",
        len(fam_ids),
        fam_path,
    )

    label_file_ids = set(gather_ids_from_csv_file(file_path=label_file))
    logger.info(
        "Gathered %d IDs from label file: %s",
        len(label_file_ids),
        label_file,
    )

    all_ids = set().union(fam_ids, label_file_ids)
    if filter_common:
        common_ids = label_file_ids.intersection(fam_ids)
        logger.info(
            "Keeping %d common IDs among %d total (difference: %d).",
            len(common_ids),
            len(all_ids),
            len(all_ids) - len(common_ids),
        )
        return list(common_ids)

    return list(all_ids)


def gather_ids_from_fam(fam_path: Path) -> set[str]:

    fam_df = pd.read_csv(
        fam_path,
        sep=r"\s+",
        header=None,
        usecols=[1],
    )
    return set(fam_df[1].astype(str))


def gather_ids_from_csv_file(
    file_path: Path,
    drop_nas: bool = False,
) -> tuple[str]:
    logger.debug("Gathering IDs from %s.", file_path)

    if drop_nas:
        df = pd.read_csv(filepath_or_buffer=file_path)
        df = df.dropna(how="any", axis=0)
    else:
        df = pd.read_csv(filepath_or_buffer=file_path, usecols=["ID"])

    all_ids = tuple(df["ID"].astype(str))

    return all_ids


def _split_ids(
    all_ids: Sequence[str],
    valid_or_test_size: float | int = 0.1,
) -> Tuple[Sequence[str], Sequence[str]]:
    assert len(all_ids) > 0
    train_ids, test_or_valid_ids = eir.data_load.label_setup.split_ids(
        ids=all_ids, valid_size=valid_or_test_size
    )

    train_ids_set = set(train_ids)
    test_ids_set = set(test_or_valid_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    return train_ids, test_or_valid_ids


def build_deeplake_train_and_test_ds_from_stream(
    train_ids: Sequence[str],
    test_ids: Sequence[str],
    source_stream: Generator[tuple[str, np.ndarray], None, None],
    destination: Path,
    chunk_size: int,
    commit_frequency: int,
) -> None:
    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    train_path = destination / "train"
    test_path = destination / "test"

    if deeplake.exists(str(train_path)) and deeplake.exists(str(test_path)):
        return

    ensure_path_exists(path=train_path, is_folder=True)
    ensure_path_exists(path=test_path, is_folder=True)

    ds_train = deeplake.create(str(train_path))
    ds_test = deeplake.create(str(test_path))

    try:
        first_id, first_array = next(source_stream)
        array_shape = list(first_array.shape)
    except StopIteration:
        raise ValueError("Source stream is empty")

    for ds in (ds_train, ds_test):
        ds.add_column("ID", dtype=deeplake.types.Text())
        array_schema = deeplake.types.Array(dtype="bool", shape=array_shape)
        ds.add_column("genotype", dtype=array_schema)
        ds.commit("Created schema")

    batch_size = chunk_size // 2
    train_batch = {"ID": [], "genotype": []}
    test_batch = {"ID": [], "genotype": []}

    if first_id in train_ids_set:
        train_batch["ID"].append(first_id)
        train_batch["genotype"].append(first_array)
    elif first_id in test_ids_set:
        test_batch["ID"].append(first_id)
        test_batch["genotype"].append(first_array)

    total_processed = 1
    skipped_samples = 0

    try:
        for id_, array in source_stream:
            if id_ in train_ids_set:
                train_batch["ID"].append(id_)
                train_batch["genotype"].append(array)

                if len(train_batch["ID"]) >= batch_size:
                    ds_train.append(train_batch)
                    train_batch = {"ID": [], "genotype": []}

            elif id_ in test_ids_set:
                test_batch["ID"].append(id_)
                test_batch["genotype"].append(array)

                if len(test_batch["ID"]) >= batch_size:
                    ds_test.append(test_batch)
                    test_batch = {"ID": [], "genotype": []}
            else:
                skipped_samples += 1

            total_processed += 1

            if total_processed % commit_frequency == 0:
                if train_batch["ID"]:
                    ds_train.append(train_batch)
                    train_batch = {"ID": [], "genotype": []}
                if test_batch["ID"]:
                    ds_test.append(test_batch)
                    test_batch = {"ID": [], "genotype": []}

                ds_train.commit(f"Processed {total_processed} samples")
                ds_test.commit(f"Processed {total_processed} samples")

                logger.info(
                    "Iterated over %d samples while splitting into train and test. "
                    "Skipped %d samples.",
                    total_processed,
                    skipped_samples,
                )

        if train_batch["ID"]:
            ds_train.append(train_batch)
        if test_batch["ID"]:
            ds_test.append(test_batch)

        ds_train.commit(f"Completed processing {total_processed} samples")
        ds_test.commit(f"Completed processing {total_processed} samples")

    except Exception as e:
        ds_train.rollback()
        ds_test.rollback()
        raise RuntimeError(f"Error splitting dataset: {str(e)}") from e


@dataclass
class SaveArrayConfig:
    id_: str
    array: np.ndarray
    base_path: Path
    is_train: bool


def _save_single_array(config: SaveArrayConfig) -> None:
    save_path = config.base_path / ("train" if config.is_train else "test")
    np.save(file=save_path / f"{config.id_}.npy", arr=config.array)


def _build_disk_array_folders_from_stream(
    train_ids: Sequence[str],
    test_ids: Sequence[str],
    source_stream: Generator[tuple[str, np.ndarray], None, None],
    destination: Path,
    batch_size: int = 1000,
    max_workers: int = 16,
) -> None:
    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)
    assert len(train_ids_set.intersection(test_ids_set)) == 0

    train_path = destination / "train"
    test_path = destination / "test"
    if train_path.exists() and test_path.exists():
        return

    train_path.mkdir(parents=True, exist_ok=True)
    test_path.mkdir(parents=True, exist_ok=True)

    cpu_count = os.cpu_count() or 1
    max_workers = min(cpu_count * 2, max_workers)

    save_fn = partial(_save_single_array)

    with ProcessPoolExecutor(max_workers=max_workers) as executor:
        batch = []

        for id_, array in source_stream:
            config = SaveArrayConfig(
                id_=id_,
                array=array,
                base_path=destination,
                is_train=id_ in train_ids_set,
            )

            batch.append(config)

            if len(batch) >= batch_size:
                futures = list(executor.map(save_fn, batch))
                _ = [f for f in futures]
                batch = []

        if batch:
            futures = list(executor.map(save_fn, batch))
            _ = [f for f in futures]


def _split_csv_into_train_and_test(
    train_ids: Sequence[str],
    test_ids: Sequence[str],
    source: Path,
    destination: Path,
) -> None:

    logger.info("Splitting label %s file into train and test sets.", source)

    ensure_path_exists(path=destination, is_folder=True)
    df_labels = pd.read_csv(filepath_or_buffer=source)
    df_labels["ID"] = df_labels["ID"].astype(str)

    train_ids_set = set(train_ids)
    test_ids_set = set(test_ids)

    df_train = df_labels[df_labels["ID"].isin(train_ids_set)].copy()
    df_test = df_labels[df_labels["ID"].isin(test_ids_set)].copy()

    df_train.to_csv(destination / "labels_train.csv", index=False)
    df_test.to_csv(destination / "labels_test.csv", index=False)


class ParseDataWrapper(luigi.Task):
    data_config = luigi.DictParameter()

    def requires(self):
        yield CommonSplitIntoTestSet(**self.data_config)

    def output(self):
        return self.input()


def get_dynamic_valid_size(
    num_samples_per_epoch: int,
    minimum: int,
    valid_size_upper_bound: int = 20000,
) -> int:
    valid_size = int(0.1 * num_samples_per_epoch)

    if valid_size < minimum:
        valid_size = minimum

    if valid_size > valid_size_upper_bound:
        valid_size = valid_size_upper_bound

    return valid_size


def get_batch_size(
    samples_per_epoch: int,
    upper_bound: int = 64,
    lower_bound: int = 4,
) -> int:
    batch_size = 2 ** int(math.log2(samples_per_epoch / 40))

    if batch_size > upper_bound:
        return upper_bound
    elif batch_size < lower_bound:
        return lower_bound

    logger.info("Batch size set to: %d", batch_size)

    if batch_size <= 8:
        logger.warning(
            "Computed batch size based on number of training"
            " samples per epoch (%d)"
            " is very small (%d). This may cause issues with training."
            " This is likely due to a small number of samples in the dataset."
            " Consider increasing the number of samples in the dataset if possible.",
            samples_per_epoch,
            batch_size,
        )

    return batch_size
